<?php

// Exit if accessed directly
if( ! defined( 'ABSPATH' ) ) exit;

/**
 * Convert amount from a currency to another
 *
 * @param $amount
 * @param $from
 * @param $to
 * @param $date
 * @return false|float
 */
function pms_convert_currency( $amount, $from, $to, $date = '' ){

    if( empty( $amount ) || empty( $from ) || empty( $to ) )
        return false;

    $mc_rates = get_option( 'pms_currency_exchange_data', array() );

    if ( !empty( $date ) ) {

        if ( isset( $mc_rates[$from][$to][$date] )  ) {
            $rate = $mc_rates[$from][$to][$date]['rate'];
        }
        else {
            $active_converter = pms_get_active_currency_converter();
            $converter = new $active_converter( $amount, $from, $to, $date );
            $rate = $converter->get_conversion_rate();
        }

    }
    else {

        if ( isset( $mc_rates[$from][$to] )  ) {
            krsort( $mc_rates[$from][$to] );
            $latest_rate = reset( $mc_rates[$from][$to] );
            $rate = $latest_rate['rate'];
        }
        else {
            $active_converter = pms_get_active_currency_converter();
            $converter = new $active_converter( $amount, $from, $to );
            $rate = $converter->get_conversion_rate();
        }

    }

    return round($amount * $rate );
}

/**
 * Get the active Converter
 *
 * @return mixed|string
 */
function pms_get_active_currency_converter(){
    $payment_settings = get_option( 'pms_payments_settings', array() );

    $convertor = isset( $payment_settings['convertor_api_provider'] ) ? $payment_settings['convertor_api_provider'] : '';

    return $convertor;
}

/**
 * Update currency conversion rates
 *
 * @return void
 */
function pms_currency_exchange_data_update_request() {
    $payment_settings = get_option( 'pms_payments_settings', array() );

    $default_currency = isset( $payment_settings['currency'] ) ? $payment_settings['currency'] : '';
    $other_currencies = isset( $payment_settings['multiple-currencies'] ) ? $payment_settings['multiple-currencies'] : array();

    if ( !empty( $default_currency ) && !empty( $other_currencies ) ) {
        $active_converter = pms_get_active_currency_converter();
        $current_date = date('Y-m-d');
        $rate_exists = true;

        foreach ( $other_currencies as $currency ) {
            $converter = new $active_converter( '1', $currency, $default_currency, $current_date );
            $rate = $converter->get_conversion_rate();

            if ( !empty( $rate ) )
                $converter->save_data( $currency, $default_currency, $rate, $current_date );
            else $rate_exists = false;

            // reverse conversion
            $converter = new $active_converter( '1', $default_currency, $currency, $current_date );
            $rate = $converter->get_conversion_rate();

            if ( !empty( $rate ) )
                $converter->save_data( $default_currency, $currency, $rate, $current_date );
            else $rate_exists = false;

        }

        if ( $rate_exists ) {
            $response['type'] = 'success';
            $response['text'] = date('Y-m-d H:i:s');
        }
        else {
            $response['type'] = 'error';
            $response['text'] = esc_html__( 'Converter limit reached or unsupported Extra Currencies selected.', 'paid-member-subscriptions' );
        }

        update_option( 'pms_currency_exchange_request_date', $response['text'], false );
    }

    if ( !wp_doing_cron() && defined('DOING_AJAX') && DOING_AJAX && isset( $response ) ) {
        wp_send_json( $response );
        wp_die();
    }

}
add_action( 'pms_multiple_currencies_cron', 'pms_currency_exchange_data_update_request' );
add_action( 'wp_ajax_pms_mc_data_update_request', 'pms_currency_exchange_data_update_request' );

/**
 * Check if conversion rate exists for the selected currency
 *
 * @param $currency
 * @return bool
 */
function pms_currency_rates_exist( $currency ) {
    $mc_rates = get_option( 'pms_currency_exchange_data', array() );

    return isset( $mc_rates[ $currency ] );
}

/**
 * Update Cron scheduled event trigger interval
 *
 * @param $old_value
 * @param $new_value
 * @return void
 */
function pms_update_cron_event( $old_value, $new_value ) {

    if (
        isset( $old_value['conversion_update_interval'] ) && $new_value['conversion_update_interval'] == $old_value['conversion_update_interval'] &&
        isset( $old_value['convertor_api_provider'] ) && $new_value['convertor_api_provider'] == $old_value['convertor_api_provider'] &&
        isset( $old_value['exchange_api_key'] ) && $new_value['exchange_api_key'] == $old_value['exchange_api_key'] &&
        isset( $old_value['alpha_vantage_api_key'] ) && $new_value['alpha_vantage_api_key'] == $old_value['alpha_vantage_api_key']
    ) {
        return;
    }

    $payment_settings = get_option( 'pms_payments_settings', array() );

    if (
        empty( $payment_settings['convertor_api_provider'] ) ||
        ( $payment_settings['convertor_api_provider'] === 'PMS_Exchange_Currency_Converter' && empty( $payment_settings['exchange_api_key'] ) ) ||
        ( $payment_settings['convertor_api_provider'] === 'PMS_Alpha_Vantage_Currency_Converter' && empty( $payment_settings['alpha_vantage_api_key'] ) )
    ) {
        wp_clear_scheduled_hook( 'pms_multiple_currencies_cron' );
        return;
    }


    if ( wp_next_scheduled( 'pms_multiple_currencies_cron' ) ) {
        wp_clear_scheduled_hook( 'pms_multiple_currencies_cron' );
    }

    wp_schedule_event( time(), $new_value['conversion_update_interval'], 'pms_multiple_currencies_cron' );

}
add_action('update_option_pms_payments_settings', 'pms_update_cron_event', 10, 2);

