<?php

/**
 * Check for PMS Free version number and Pro Rate Add-On state before:
 * - activating Multiple Currencies Add-On
 * - schedule Cron event for conversion rates update
 *
 * @param $slug
 * @param $extra_data
 * @return void
 */
function pms_mc_schedule_cron( $slug, $extra_data = false ) {

    $possible_slugs = array(
        'pms-add-on-multiple-currencies/index.php',
        'paid-member-subscriptions-pro/index.php',
        'paid-member-subscriptions-unlimited/index.php',
        'paid-member-subscriptions-agency/index.php',
        'paid-member-subscriptions-basic/index.php',
        'paid-member-subscriptions-dev/index.php',
    );

    if ( !in_array( $slug, $possible_slugs ) )
        return;

    // don't activate if PMS Free version number smaller than 2.13.5 or Pro Rate Add-On is active
    $pms_add_ons_settings = get_option( 'pms_add_ons_settings', array() );
    $pro_rate_add_on_active = !empty( $pms_add_ons_settings['pms-add-on-pro-rate/index.php'] ) ? $pms_add_ons_settings['pms-add-on-pro-rate/index.php'] : false;

    if( !version_compare( PMS_VERSION, '2.13.5', '>=' ) ){
        $activation_failed_data['message'] = esc_html__( 'The version of Paid Member Subscriptions you are using is not compatible with the Multiple Currencies add-on. Please update Paid Member Subscriptions to version 2.13.5 or higher.', 'paid-member-subscriptions' );
        $activation_failed_data['failure_reason'] = 'base_version_incompatibility';
    }
    elseif( $pro_rate_add_on_active ){
        $activation_failed_data['message'] = esc_html__( 'The Multiple Currencies Add-On is not compatible with the Pro-Rate Add-On. To use Multiple Currencies, please deactivate the Pro-Rate Add-On.', 'paid-member-subscriptions' );
        $activation_failed_data['failure_reason'] = 'pro_rate_incompatibility';
    }

    if ( isset( $activation_failed_data['message'] ) ) {
        $activation_failed_data['failed_activation'] = true;
        update_option( 'pms_mc_activation_failed', $activation_failed_data );

        $pms_add_ons_settings['pms-add-on-multiple-currencies/index.php'] = false;
        update_option( 'pms_add_ons_settings', $pms_add_ons_settings );

        return;
    }
    else delete_option( 'pms_mc_activation_failed' );


    $payment_settings = get_option( 'pms_payments_settings', array() );

    // set default Conversion Type
    if ( !isset( $payment_settings['conversion_type'] ) ) {
        $payment_settings['conversion_type'] = 'manual';
        update_option( 'pms_payments_settings', $payment_settings );
    }

    // don't schedule Cron event if Converter or API Key (if needed) is missing
    if (
        empty( $payment_settings['convertor_api_provider'] ) ||
        ( $payment_settings['convertor_api_provider'] === 'PMS_Exchange_Currency_Converter' && empty( $payment_settings['exchange_api_key'] ) ) ||
        ( $payment_settings['convertor_api_provider'] === 'PMS_Alpha_Vantage_Currency_Converter' && empty( $payment_settings['alpha_vantage_api_key'] ) )
    ) {
        return;
    }

    if ( !wp_next_scheduled( 'pms_multiple_currencies_cron' ) && isset( $payment_settings['conversion_update_interval'] ) )
        wp_schedule_event( time(), $payment_settings['conversion_update_interval'], 'pms_multiple_currencies_cron' );

}
add_action( 'pms_add_ons_activate', 'pms_mc_schedule_cron' );
add_action( 'activated_plugin', 'pms_mc_schedule_cron', 10, 2 );


/**
 * Handle Multiple Currencies Add-On incompatibility admin notice
 *
 * @return void
 */
function pms_mc_compatibility_notice() {

    if ( defined('DOING_AJAX') && DOING_AJAX )
        return;

    $failed_activation = get_option( 'pms_mc_activation_failed', false );

    if ( !$failed_activation )
        return;

    $pms_add_ons_settings = get_option( 'pms_add_ons_settings', array() );
    $pro_rate_add_on_active = !empty( $pms_add_ons_settings['pms-add-on-pro-rate/index.php'] ) ? $pms_add_ons_settings['pms-add-on-pro-rate/index.php'] : false;
    $multiple_currencies_add_on_active = !empty( $pms_add_ons_settings['pms-add-on-multiple-currencies/index.php'] ) ? $pms_add_ons_settings['pms-add-on-multiple-currencies/index.php'] : false;

    if (
        ( $failed_activation['failure_reason'] == 'pro_rate_incompatibility' && !$pro_rate_add_on_active && !$multiple_currencies_add_on_active ) ||
        ( $failed_activation['failure_reason'] == 'base_version_incompatibility' && version_compare( PMS_VERSION, '2.13.5', '>=' ) )
    ) {
        delete_option( 'pms_mc_activation_failed' );
        return;
    }

    if ( !empty( $failed_activation['message'] ) ) {
        $pms_notifications_instance = PMS_Plugin_Notifications::get_instance();
        $pms_notifications_instance->add_notification( 'pms_multiple_currencies_incompatibility', $failed_activation['message'], 'pms-notice error notice is-dismissible', false );
    }

}
add_action( 'admin_init', 'pms_mc_compatibility_notice' );


/**
 * Remove Cron event
 *
 * @param $slug
 * @param $extra_data
 * @return void
 */
function pms_mc_clear_scheduled_cron( $slug, $extra_data = false ) {

    $possible_slugs = array(
        'pms-add-on-multiple-currencies/index.php',
        'paid-member-subscriptions-pro/index.php',
        'paid-member-subscriptions-unlimited/index.php',
        'paid-member-subscriptions-agency/index.php',
        'paid-member-subscriptions-basic/index.php',
        'paid-member-subscriptions-dev/index.php',
    );

    if ( !in_array( $slug, $possible_slugs ) )
        return;

    wp_clear_scheduled_hook( 'pms_multiple_currencies_cron' );

}
add_action( 'pms_add_ons_deactivate','pms_mc_clear_scheduled_cron' );
add_action('deactivated_plugin', 'pms_mc_clear_scheduled_cron', 10, 2);


/**
 * Add custom interval for Cron scheduled event
 *
 * @param $schedules
 * @return mixed
 */
function pms_add_custom_cron_schedules( $schedules ) {
    if ( !isset( $schedules['every_6_hours'] ) ) {
        $schedules['every_6_hours'] = array(
            'interval' => 6 * HOUR_IN_SECONDS,
            'display'  => esc_html__('Every 6 Hours', 'paid-member-subscriptions'),
        );
    }
    return $schedules;
}
add_filter( 'cron_schedules', 'pms_add_custom_cron_schedules' );